import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;



class GenJava implements Visitor<Void>{
  ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
  
  MethodVisitor mv = null;  

  Map<String,Integer> env = new HashMap<>();
  Map<String,Tree.Param> vars = new HashMap<>();
  Map<String,Tree.FunDef> funs = new HashMap<>(); 
  Map<String,Tree.ConDef> cons = new HashMap<>(); 

  
  @Override public Void visit(Tree.FunDef fd){
    var args = "";
    for (var a:fd.args) args+="I";
    mv = cw.visitMethod(Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC, fd.name, "("+args+")I", null, null);
    var i = 0;
    for (var a:fd.args) {
    	env.put(a.name,i++);
    	vars.put(a.name,a);
    }
    fd.body.welcome(this);
    mv.visitInsn(Opcodes.IRETURN);
    mv.visitMaxs(1, 1);    
    mv.visitEnd();
    return null;
  }
  
  @Override public Void visit(Tree.Prog prog){
	for (var theType:prog.types) theType.welcome(this);
	  
	for (var fun:prog.funs) funs.put(fun.name,fun);
	for (var type:prog.types)for(var con:type.cons) cons.put(con.name,con);
	
    cw.visit(Opcodes.V1_8, Opcodes.ACC_PUBLIC + Opcodes.ACC_SUPER, "Module", null, "java/lang/Object", null);
    cw.visitSource("Module.java", null);

    for(var fundef:prog.funs) fundef.welcome(this);

    mv = cw.visitMethod(Opcodes.ACC_PUBLIC | Opcodes.ACC_STATIC, "main", "([Ljava/lang/String;)V", null, null);
    mv.visitFieldInsn(Opcodes.GETSTATIC, "java/lang/System", "out", "Ljava/io/PrintStream;");

    prog.expr.welcome(this);

/*	if (convertType(prog.expr.theType).equals("I"))
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(I)V", false);
	else
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/Object;)V", false);*/

    
    mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/io/PrintStream", "println", "(Ljava/lang/Object;)V", false);
    mv.visitInsn(Opcodes.RETURN);
    mv.visitMaxs(1, 1);
    
    mv.visitEnd();

    cw.visitEnd();

    try {
      OutputStream out = new FileOutputStream("Module.class");
      out.write(cw.toByteArray());
      out.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
    return null;
  }
  @Override public Void visit(Tree.MultExpr e){
    e.op1.welcome(this);
    e.op2.welcome(this);
    mv.visitInsn(Opcodes.IMUL);
    return null;
  }
  @Override public Void visit(Tree.AddExpr e){
    e.op1.welcome(this);
    e.op2.welcome(this);
    mv.visitInsn(Opcodes.IADD);

    return null;
  }

  @Override public Void visit(Tree.IntLit e){
    mv.visitLdcInsn(e.n);
    return null;
  }

  public static void main(String[] args) throws Exception{
    LParser parser = new LParser(new FileReader(args[0]));
    var gj = new GenJava();
    System.out.println(parser.program().welcome(gj));

  }

@Override
public Void visit(Tree.Var e) {
	if (vars.get(e.name).theType.equals("Int"))
		mv.visitVarInsn(Opcodes.ILOAD, env.get(e.name));
	else 
		mv.visitVarInsn(Opcodes.ALOAD, env.get(e.name));

	return null;
}

@Override
public Void visit(Tree.FunCall e) {
    var fd = funs.get(e.name);
    if (fd!=null) {
    	var args="";
    	e.args.forEach(arg -> arg.welcome(this));
    	for (var a:fd.args) args+=vmType(a.theType);
    	mv.visitMethodInsn(Opcodes.INVOKESTATIC,"Module", e.name, "("+args+")I", false);
    }else {
        var con = cons.get(e.name);
        if (con==null) throw new RuntimeException("Unknown Function name: "+e.name);
        mv.visitTypeInsn(Opcodes.NEW, con.name);
		mv.visitInsn(Opcodes.DUP);
		e.args.forEach(arg -> arg.welcome(this));
		StringBuffer signature = new StringBuffer("(");
		con.args.forEach(arg -> signature.append(vmType(arg.theType)));
		signature.append(")");
		signature.append("V");
		mv.visitMethodInsn(Opcodes.INVOKESPECIAL, con.name, "<init>", signature + "", false);
    }
	
	return null;
}

@Override
public Void visit(Tree.TypeDef typeDef) {
	ClassWriter tw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
    tw.visit(Opcodes.V1_8, Opcodes.ACC_PUBLIC + Opcodes.ACC_ABSTRACT + Opcodes.ACC_INTERFACE, 
    		typeDef.name, null, "java/lang/Object", null);
    cw.visitSource(typeDef.name+".java", null);
    cw.visitEnd();

    try {
      OutputStream out = new FileOutputStream(typeDef.name+".class");
      out.write(tw.toByteArray());
      out.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }

    for (var con:typeDef.cons) {
    	con.typeName = typeDef.name;
    	con.welcome(this);
    }
    
	return null;
}

private static String vmType(String type) {
	switch (type) {
	case "Int": return "I";
	default: return "L"+type+";";
	}
}

@Override
public Void visit(Tree.ConDef conDef) {
	ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
    cw.visit(Opcodes.V1_8, Opcodes.ACC_PUBLIC , 
    		conDef.name, null, "java/lang/Object", new String[] {conDef.typeName});
    cw.visitSource(conDef.name+".sourcelang", null);

    for (var arg:conDef.args) {
    	cw.visitField(Opcodes.ACC_PUBLIC,arg.name,vmType(arg.theType),null,null);
    	
    }
    String signature = "(";
	for (var arg : conDef.args) {
		signature += vmType(arg.theType);
	}
    MethodVisitor mv = cw.visitMethod(Opcodes.ACC_PUBLIC, "<init>", signature + ")V", null, null);   
    mv.visitVarInsn(Opcodes.ALOAD, 0);
    var i = 1;
	mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "java/lang/Object", "<init>", "()V", false);
	for (var arg : conDef.args) {
		mv.visitVarInsn(Opcodes.ALOAD, 0);
		if (arg.theType.equals("Int")) {
			mv.visitVarInsn(Opcodes.ILOAD, i);
		} else {
			mv.visitVarInsn(Opcodes.ALOAD, i);
		}
		mv.visitFieldInsn(Opcodes.PUTFIELD, conDef.name, arg.name, vmType(arg.theType));
		i++;
	}
	mv.visitInsn(Opcodes.RETURN);
	mv.visitMaxs(1, 1);
	mv.visitEnd();
    
	generateToString(cw, conDef);
	
    cw.visitEnd();

    try {
      OutputStream out = new FileOutputStream(conDef.name+".class");
      out.write(cw.toByteArray());
      out.close();
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
	return null;
}

@Override
public Void visit(Tree.Param param) {
	// TODO Auto-generated method stub
	return null;
}

private void generateToString(ClassWriter cv, Tree.ConDef conDef) {
	MethodVisitor mv = cv.visitMethod(Opcodes.ACC_PUBLIC, "toString", "()Ljava/lang/String;", null, null);
	mv.visitTypeInsn(Opcodes.NEW, "java/lang/StringBuffer");
	mv.visitInsn(Opcodes.DUP);
	mv.visitMethodInsn(Opcodes.INVOKESPECIAL, "java/lang/StringBuffer", "<init>", "()V", false);
	mv.visitVarInsn(Opcodes.ASTORE, 1);
	mv.visitVarInsn(Opcodes.ALOAD, 1);
	mv.visitLdcInsn(conDef.name);
	mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
			"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
	mv.visitLdcInsn("(");
	mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
			"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
	int i = 1;
	for (var arg : conDef.args) {
		if (i > 1) {
			mv.visitLdcInsn(", ");
			mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
					"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
		}
		boolean prim = arg.theType.equals("Int");
		mv.visitVarInsn(Opcodes.ALOAD, 0);
		String typ = prim ? "I" : "Ljava/lang/Object;";
		mv.visitFieldInsn(Opcodes.GETFIELD, conDef.name, arg.name, vmType(arg.theType));
		mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
				"(" + typ + ")Ljava/lang/StringBuffer;", false);

		i++;
	}
	mv.visitLdcInsn(")");
	mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "append",
			"(Ljava/lang/String;)Ljava/lang/StringBuffer;", false);
	mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "java/lang/StringBuffer", "toString", "()Ljava/lang/String;", false);
	mv.visitInsn(Opcodes.ARETURN);
	mv.visitMaxs(1, 1);
	mv.visitEnd();

}


}
