import java.util.List;

interface Tree {
	<R> R welcome(Visitor<R> v);

	static class Prog implements Tree {
		List<FunDef> funs;
		List<TypeDef> types;
		Tree expr;

		Prog(List<TypeDef> types,List<FunDef> funs, Tree expr) {
			this.funs = funs;
			this.expr = expr;
			this.types = types;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class TypeDef implements Tree {
		String name;
		List<ConDef> cons;

		public TypeDef(String name, List<ConDef> cons) {
			super();
			this.name = name;
			this.cons = cons;
		}


		@Override 
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class ConDef implements Tree {
		String name;
		List<Param> args;
		String typeName = null;

		 
		public ConDef(String name, List<Param> args) {
			super();
			this.name = name;
			this.args = args;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class Param implements Tree {
		String name;
		String theType;
		
		public Param(String name, String theType) {
			super();
			this.name = name;
			this.theType = theType;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class FunDef implements Tree {
		String name;
		List<Param> args;
		Tree body;

		FunDef(String name, List<Param> args, Tree body) {
			this.name = name;
			this.args = args;
			this.body = body;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class FunCall implements Tree {
		String name;
		List<Tree> args;

		FunCall(String name, List<Tree> args) {
			this.name = name;
			this.args = args;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class Var implements Tree {
		String name;

		Var(String name) {
			this.name = name;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class IntLit implements Tree {
		int n;

		IntLit(int n) {
			this.n = n;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}
	}

	static class MultExpr implements Tree {
		Tree op1;
		Tree op2;

		MultExpr(Tree op1, Tree op2) {
			this.op1 = op1;
			this.op2 = op2;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}

	}

	static class AddExpr implements Tree {
		Tree op1;
		Tree op2;

		AddExpr(Tree op1, Tree op2) {
			this.op1 = op1;
			this.op2 = op2;
		}

		@Override
		public <R> R welcome(Visitor<R> v) {
			return v.visit(this);
		}

	}

}
